import json
import sys
from typing import Any


class Validator:
    VALID_LEVELS = {"debug", "info", "warning", "error", "critical"}
    VALID_LOG_TYPES = {"console", "logger", "exception", "custom"}
    VALID_IMPORTANCE = {"low", "standard", "high"}

    def __init__(self, constraints: dict[str, Any]):
        self.max_message_length = constraints.get("max_message_length", 10000)
        self.max_error_message_length = constraints.get("max_error_message_length", 5000)
        self.max_stack_trace_length = constraints.get("max_stack_trace_length", 50000)
        self.max_attributes_size_bytes = constraints.get("max_attributes_size_bytes", 102400)

    def validate_log(self, log_entry: dict[str, Any]) -> dict[str, Any]:
        validated = log_entry.copy()

        if "level" not in validated or validated["level"] not in self.VALID_LEVELS:
            validated["level"] = "info"

        if "log_type" not in validated or validated["log_type"] not in self.VALID_LOG_TYPES:
            validated["log_type"] = "console"

        if "importance" not in validated or validated["importance"] not in self.VALID_IMPORTANCE:
            validated["importance"] = "standard"

        if validated.get("message"):
            validated["message"] = self._truncate_string(
                validated["message"], self.max_message_length, "message"
            )

        if validated.get("error_message"):
            validated["error_message"] = self._truncate_string(
                validated["error_message"], self.max_error_message_length, "error_message"
            )

        if validated.get("stack_trace"):
            validated["stack_trace"] = self._truncate_string(
                validated["stack_trace"], self.max_stack_trace_length, "stack_trace"
            )

        if validated.get("attributes"):
            validated["attributes"] = self._validate_attributes(validated["attributes"])

        if "timestamp" in validated:
            validated["timestamp"] = self._normalize_timestamp(validated["timestamp"])

        return validated

    def _truncate_string(self, value: str, max_length: int, field_name: str) -> str:
        if not isinstance(value, str):
            value = str(value)

        if len(value) <= max_length:
            return value

        truncated_suffix = "... [truncated]"
        truncate_at = max_length - len(truncated_suffix)

        self._log_validation_warning(
            f"Field '{field_name}' truncated from {len(value)} to {max_length} characters"
        )

        return value[:truncate_at] + truncated_suffix

    def _validate_attributes(self, attributes: Any) -> dict[str, Any]:
        if not isinstance(attributes, dict):
            self._log_validation_warning(
                f"Attributes must be a dict, got {type(attributes).__name__}, converting"
            )
            return {"value": str(attributes)}

        try:
            serialized = json.dumps(attributes)
            size_bytes = len(serialized.encode("utf-8"))

            if size_bytes > self.max_attributes_size_bytes:
                self._log_validation_warning(
                    f"Attributes size ({size_bytes} bytes) exceeds max ({self.max_attributes_size_bytes} bytes), truncating"
                )

                return self._truncate_attributes(attributes, self.max_attributes_size_bytes)

            return attributes

        except (TypeError, ValueError) as e:
            self._log_validation_warning(f"Attributes not JSON serializable: {e}, removing")
            return {}

    def _truncate_attributes(self, attributes: dict[str, Any], max_bytes: int) -> dict[str, Any]:
        result = {}
        current_bytes = 2

        for key, value in attributes.items():
            try:
                item_json = json.dumps({key: value})
                item_bytes = len(item_json.encode("utf-8"))

                if current_bytes + item_bytes <= max_bytes - 100:
                    result[key] = value
                    current_bytes += item_bytes
                else:
                    result["_truncated"] = True
                    break

            except (TypeError, ValueError):
                continue

        return result

    def _normalize_timestamp(self, timestamp: str) -> str:
        if isinstance(timestamp, str) and timestamp.endswith("Z"):
            return timestamp

        if isinstance(timestamp, str):
            return timestamp

        return str(timestamp)

    def _log_validation_warning(self, message: str) -> None:
        sys.stderr.write(f"[Ledger SDK] [VALIDATION WARNING] {message}\n")
        sys.stderr.flush()



import pytest

from ledger.core.validator import Validator


class TestValidator:
    @pytest.fixture
    def validator(self):
        constraints = {
            "max_message_length": 100,
            "max_error_message_length": 50,
            "max_stack_trace_length": 500,
            "max_attributes_size_bytes": 1024,
        }
        return Validator(constraints)

    def test_validate_valid_log(self, validator):
        log_entry = {
            "timestamp": "2025-01-11T10:00:00Z",
            "level": "info",
            "log_type": "console",
            "importance": "standard",
            "message": "Test message",
        }

        validated = validator.validate_log(log_entry)

        assert validated["level"] == "info"
        assert validated["log_type"] == "console"
        assert validated["importance"] == "standard"
        assert validated["message"] == "Test message"

    def test_validate_with_invalid_level(self, validator):
        log_entry = {
            "level": "invalid_level",
            "log_type": "console",
            "importance": "standard",
        }

        validated = validator.validate_log(log_entry)

        assert validated["level"] == "info"

    def test_truncate_long_message(self, validator):
        long_message = "A" * 200
        log_entry = {
            "level": "info",
            "log_type": "console",
            "importance": "standard",
            "message": long_message,
        }

        validated = validator.validate_log(log_entry)

        assert len(validated["message"]) == 100
        assert validated["message"].endswith("... [truncated]")

    def test_validate_attributes(self, validator):
        log_entry = {
            "level": "info",
            "log_type": "console",
            "importance": "standard",
            "attributes": {"user_id": 123, "action": "login"},
        }

        validated = validator.validate_log(log_entry)

        assert validated["attributes"]["user_id"] == 123
        assert validated["attributes"]["action"] == "login"

    def test_validate_invalid_attributes(self, validator):
        log_entry = {
            "level": "info",
            "log_type": "console",
            "importance": "standard",
            "attributes": "not_a_dict",
        }

        validated = validator.validate_log(log_entry)

        assert isinstance(validated["attributes"], dict)
        assert "value" in validated["attributes"]

    def test_normalize_timestamp(self, validator):
        log_entry = {
            "level": "info",
            "log_type": "console",
            "importance": "standard",
            "timestamp": "2025-01-11T10:00:00Z",
        }

        validated = validator.validate_log(log_entry)

        assert validated["timestamp"] == "2025-01-11T10:00:00Z"



from typing import Any


class SettingsManager:
    def __init__(self):
        self._settings = self._get_default_settings()

    def get_settings(self) -> dict[str, Any]:
        return self._settings

    def get_rate_limits(self) -> dict[str, int]:
        return self._settings["rate_limits"]

    def get_constraints(self) -> dict[str, Any]:
        return self._settings["constraints"]

    def _get_default_settings(self) -> dict[str, Any]:
        return {
            "rate_limits": {
                "requests_per_minute": 1000,
                "requests_per_hour": 50000,
            },
            "constraints": {
                "max_batch_size": 1000,
                "max_message_length": 10000,
                "max_error_message_length": 5000,
                "max_stack_trace_length": 50000,
                "max_attributes_size_bytes": 102400,
            },
        }



from unittest.mock import AsyncMock, Mock

import pytest


@pytest.fixture
def mock_http_client():
    client = Mock()
    client.post = AsyncMock()
    client.get = AsyncMock()
    client.close = AsyncMock()
    return client


@pytest.fixture
def api_key():
    return "ledger_proj_1_test_key_12345"


@pytest.fixture
def base_url():
    return "http://localhost:8000"



from typing import Any

import httpx


class HTTPClient:
    def __init__(
        self,
        base_url: str,
        api_key: str,
        timeout: float = 5.0,
        pool_size: int = 10,
    ):
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key

        self._client = httpx.AsyncClient(
            base_url=self.base_url,
            timeout=httpx.Timeout(timeout),
            limits=httpx.Limits(
                max_connections=pool_size,
                max_keepalive_connections=pool_size,
                keepalive_expiry=30.0,
            ),
            headers={
                "Authorization": f"Bearer {self.api_key}",
                "Content-Type": "application/json",
                "User-Agent": "ledger-sdk-python/1.0.0",
            },
        )

    async def post(
        self,
        path: str,
        json_data: dict[str, Any],
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        merged_headers = self._client.headers.copy()
        if headers:
            merged_headers.update(headers)

        response = await self._client.post(
            path,
            json=json_data,
            headers=merged_headers,
        )
        return response

    async def get(
        self,
        path: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> httpx.Response:
        merged_headers = self._client.headers.copy()
        if headers:
            merged_headers.update(headers)

        response = await self._client.get(
            path,
            params=params,
            headers=merged_headers,
        )
        return response

    async def close(self) -> None:
        await self._client.aclose()



import os
import sys
from contextlib import asynccontextmanager
from pathlib import Path

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from dotenv import load_dotenv
from fastapi import FastAPI, HTTPException
from ledger import LedgerClient
from ledger.integrations.fastapi import LedgerMiddleware

env_path = Path(__file__).parent.parent / ".env"
if env_path.exists():
    load_dotenv(env_path)
    print(f"[OK] Loaded configuration from {env_path}")
else:
    print(f"[WARNING] No .env file found at {env_path}")
    print("  Run: python scripts/setup_test_account.py")

api_key = os.getenv("LEDGER_API_KEY")
base_url = os.getenv("LEDGER_BASE_URL")

if not api_key:
    print("\n" + "=" * 60)
    print("ERROR: LEDGER_API_KEY not set!")
    print("=" * 60)
    print("\nPlease run the setup script first:")
    print("  python scripts/setup_test_account.py")
    print("\nOr set the environment variable manually:")
    print("  export LEDGER_API_KEY=your_api_key")
    print()
    sys.exit(1)

ledger = LedgerClient(
    api_key=api_key,
    base_url=base_url,
    flush_interval=5.0,
    max_buffer_size=10000,
)


@asynccontextmanager
async def lifespan(app: FastAPI):
    yield
    print("Shutting down, flushing remaining logs...")
    await ledger.shutdown(timeout=10.0)
    print("Shutdown complete")


app = FastAPI(title="Ledger SDK Example", lifespan=lifespan)

app.add_middleware(
    LedgerMiddleware,
    ledger_client=ledger,
    exclude_paths=["/health", "/metrics"],
)


@app.get("/")
async def root():
    return {"message": "Hello World", "status": "ok"}


@app.get("/health")
async def health():
    return {"status": "healthy"}


@app.get("/user/{user_id}")
async def get_user(user_id: int):
    ledger.log_info(f"Fetching user {user_id}", attributes={"user_id": user_id})

    if user_id == 0:
        raise HTTPException(status_code=404, detail="User not found")

    return {"user_id": user_id, "name": f"User {user_id}"}


@app.post("/payment")
async def process_payment(amount: float):
    if amount < 0:
        raise ValueError("Amount must be positive")

    if amount > 10000:
        raise ValueError("Amount exceeds maximum limit")

    ledger.log_info(
        f"Payment processed: ${amount}",
        attributes={"amount": amount, "currency": "USD"},
    )

    return {"status": "success", "amount": amount}


@app.get("/error")
async def trigger_error():
    try:
        result = 1 / 0
    except Exception as e:
        ledger.log_exception(e, message="Division by zero error")
        raise HTTPException(status_code=500, detail="Internal server error")


@app.get("/metrics")
async def get_metrics():
    return ledger.get_metrics()


@app.get("/sdk/health")
async def sdk_health():
    return ledger.get_health_status()


if __name__ == "__main__":
    import uvicorn

    from ledger.core.config import DEFAULT_CONFIG

    print("\n" + "=" * 60)
    print("Ledger SDK Example App")
    print("=" * 60)
    print(f"Ledger Server:  {base_url or DEFAULT_CONFIG.base_url}")
    print(
        f"API Key:        {api_key[:20]}..." if len(api_key) > 20 else f"API Key:        {api_key}"
    )
    print(f"App URL:        http://localhost:8080")
    print(f"API Docs:       http://localhost:8080/docs")
    print("=" * 60)
    print("\nExample endpoints:")
    print("  GET  /              - Hello world")
    print("  GET  /user/123      - Get user (with logging)")
    print("  GET  /user/0        - Trigger 404 error")
    print('  POST /payment       - Process payment (with body: {"amount": 100})')
    print("  GET  /error         - Trigger exception")
    print("  GET  /metrics       - View SDK metrics")
    print("  GET  /health        - Health check (not logged)")
    print()

    uvicorn.run(app, host="0.0.0.0", port=8080)



import asyncio
import time
from collections import deque


class RateLimiter:
    def __init__(
        self,
        requests_per_minute: int,
        requests_per_hour: int = 50000,
        buffer: float = 0.9,
    ):
        self.limit_per_minute = int(requests_per_minute * buffer)
        self.limit_per_hour = int(requests_per_hour * buffer)
        self.buffer_ratio = buffer

        self.timestamps_minute: deque[float] = deque()
        self.timestamps_hour: deque[float] = deque()

    async def wait_if_needed(self) -> None:
        now = time.time()

        self._cleanup_timestamps(now)

        sleep_time = self._calculate_sleep_time(now)

        if sleep_time > 0:
            await asyncio.sleep(sleep_time)
            now = time.time()
            self._cleanup_timestamps(now)

        self.timestamps_minute.append(now)
        self.timestamps_hour.append(now)

    def _cleanup_timestamps(self, now: float) -> None:
        minute_ago = now - 60
        while self.timestamps_minute and self.timestamps_minute[0] < minute_ago:
            self.timestamps_minute.popleft()

        hour_ago = now - 3600
        while self.timestamps_hour and self.timestamps_hour[0] < hour_ago:
            self.timestamps_hour.popleft()

    def _calculate_sleep_time(self, now: float) -> float:
        sleep_times = []

        if len(self.timestamps_minute) >= self.limit_per_minute:
            oldest_minute = self.timestamps_minute[0]
            sleep_until_minute = oldest_minute + 60
            sleep_times.append(sleep_until_minute - now)

        if len(self.timestamps_hour) >= self.limit_per_hour:
            oldest_hour = self.timestamps_hour[0]
            sleep_until_hour = oldest_hour + 3600
            sleep_times.append(sleep_until_hour - now)

        if sleep_times:
            return max(0, max(sleep_times))

        return 0

    def get_current_rate(self) -> int:
        now = time.time()
        self._cleanup_timestamps(now)
        return len(self.timestamps_minute)

    def get_current_hourly_rate(self) -> int:
        now = time.time()
        self._cleanup_timestamps(now)
        return len(self.timestamps_hour)

    def is_at_limit(self) -> bool:
        now = time.time()
        self._cleanup_timestamps(now)
        return (
            len(self.timestamps_minute) >= self.limit_per_minute
            or len(self.timestamps_hour) >= self.limit_per_hour
        )



from pydantic import Field
from pydantic_settings import BaseSettings


class LedgerConfig(BaseSettings):

    base_url: str = Field(
        default="https://ledger-server.jtuta.cloud",
        description="Ledger server URL - single source of truth",
    )

    flush_interval: float = Field(
        default=5.0, gt=0, description="Interval in seconds between automatic buffer flushes"
    )

    flush_size: int = Field(
        default=100, gt=0, description="Number of logs that trigger an automatic flush"
    )

    max_buffer_size: int = Field(
        default=10000, gt=0, description="Maximum number of logs to keep in buffer"
    )

    http_timeout: float = Field(default=5.0, gt=0, description="HTTP request timeout in seconds")

    http_pool_size: int = Field(default=10, gt=0, description="HTTP connection pool size")

    rate_limit_buffer: float = Field(
        default=0.9, gt=0, le=1, description="Rate limit buffer percentage"
    )

    class Config:
        env_prefix = "LEDGER_"
        case_sensitive = False


DEFAULT_CONFIG = LedgerConfig()



import pytest
from ledger import LedgerClient


class TestLedgerClient:
    @pytest.mark.asyncio
    async def test_client_initialization(self, api_key, base_url):
        client = LedgerClient(api_key=api_key, base_url=base_url)

        assert client.api_key == api_key
        assert client.base_url == base_url

        await client.shutdown(timeout=0.1)

    def test_invalid_api_key(self):
        with pytest.raises(ValueError) as exc_info:
            LedgerClient(api_key="invalid_key", base_url="http://localhost:8000")

        assert "api_key must start with 'ledger_' prefix" in str(exc_info.value)

    def test_invalid_base_url(self, api_key):
        with pytest.raises(ValueError) as exc_info:
            LedgerClient(api_key=api_key, base_url="not_a_url")

        assert "base_url must start with 'http://' or 'https://'" in str(exc_info.value)

    def test_invalid_flush_interval(self, api_key, base_url):
        with pytest.raises(ValueError) as exc_info:
            LedgerClient(api_key=api_key, base_url=base_url, flush_interval=-1)

        assert "flush_interval must be positive" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_log_info(self, api_key, base_url):
        client = LedgerClient(api_key=api_key, base_url=base_url)

        client.log_info("Test message", attributes={"user_id": 123})

        assert client._buffer.size() == 1

        await client.shutdown(timeout=0.1)

    @pytest.mark.asyncio
    async def test_log_error(self, api_key, base_url):
        client = LedgerClient(api_key=api_key, base_url=base_url)

        client.log_error("Error message", attributes={"error_code": "ERR_001"})

        assert client._buffer.size() == 1

        await client.shutdown(timeout=0.1)

    @pytest.mark.asyncio
    async def test_log_exception(self, api_key, base_url):
        client = LedgerClient(api_key=api_key, base_url=base_url)

        try:
            raise ValueError("Test exception")
        except ValueError as e:
            client.log_exception(e, message="Custom message")

        assert client._buffer.size() == 1

        await client.shutdown(timeout=0.1)

    @pytest.mark.asyncio
    async def test_is_healthy(self, api_key, base_url):
        client = LedgerClient(api_key=api_key, base_url=base_url)

        assert client.is_healthy() is True

        await client.shutdown(timeout=0.1)

    @pytest.mark.asyncio
    async def test_get_health_status(self, api_key, base_url):
        client = LedgerClient(api_key=api_key, base_url=base_url)

        status = client.get_health_status()

        assert status["status"] == "healthy"
        assert status["healthy"] is True
        assert status["issues"] is None

        await client.shutdown(timeout=0.1)

    @pytest.mark.asyncio
    async def test_get_metrics(self, api_key, base_url):
        client = LedgerClient(api_key=api_key, base_url=base_url)

        metrics = client.get_metrics()

        assert "sdk" in metrics
        assert "buffer" in metrics
        assert "flusher" in metrics
        assert "rate_limiter" in metrics
        assert "errors" in metrics
        assert metrics["sdk"]["version"] == "1.0.3"

        await client.shutdown(timeout=0.1)

    @pytest.mark.asyncio
    async def test_shutdown(self, api_key, base_url):
        client = LedgerClient(api_key=api_key, base_url=base_url)

        await client.shutdown(timeout=1.0)

        assert client._buffer.is_empty() or not client._buffer.is_empty()



import time
from collections.abc import Callable

from fastapi import Request, Response
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.types import ASGIApp

import ledger.core.client as client_module


class LedgerMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        ledger_client: "client_module.LedgerClient",
        exclude_paths: list[str] | None = None,
        capture_query_params: bool = True,
    ):
        super().__init__(app)
        self.ledger = ledger_client
        self.exclude_paths = exclude_paths or []
        self.capture_query_params = capture_query_params

    async def dispatch(
        self,
        request: Request,
        call_next: Callable[[Request], Response],
    ) -> Response:
        if request.url.path in self.exclude_paths:
            return await call_next(request)

        start_time = time.time()

        request_info = {
            "method": request.method,
            "path": request.url.path,
        }

        if self.capture_query_params and request.url.query:
            request_info["query_params"] = str(request.url.query)

        try:
            response = await call_next(request)

            duration_ms = (time.time() - start_time) * 1000

            self._log_request(request_info, response.status_code, duration_ms)

            return response

        except Exception as exc:
            duration_ms = (time.time() - start_time) * 1000

            self._log_exception(request_info, exc, duration_ms)

            raise

    def _log_request(
        self,
        request_info: dict[str, str],
        status_code: int,
        duration_ms: float,
    ) -> None:
        if 200 <= status_code < 400:
            level = "info"
            importance = "standard"
        elif 400 <= status_code < 500:
            level = "warning"
            importance = "standard"
        else:
            level = "error"
            importance = "high"

        message = (
            f"{request_info['method']} {request_info['path']} - {status_code} ({duration_ms:.0f}ms)"
        )

        self.ledger._log(
            level=level,
            log_type="console",
            importance=importance,
            message=message,
            attributes={
                "http": {
                    "method": request_info["method"],
                    "path": request_info["path"],
                    "status_code": status_code,
                    "duration_ms": round(duration_ms, 2),
                    "query_params": request_info.get("query_params"),
                }
            },
        )

    def _log_exception(
        self,
        request_info: dict[str, str],
        exception: Exception,
        duration_ms: float,
    ) -> None:
        message = f"{request_info['method']} {request_info['path']} - Exception: {exception!s}"

        self.ledger.log_exception(
            exception=exception,
            message=message,
            attributes={
                "http": {
                    "method": request_info["method"],
                    "path": request_info["path"],
                    "duration_ms": round(duration_ms, 2),
                    "query_params": request_info.get("query_params"),
                }
            },
        )



import asyncio
import sys
import time
from typing import Any

import httpx

import ledger.core.buffer as buffer_module
import ledger.core.http_client as http_client_module
import ledger.core.rate_limiter as rate_limiter_module


class BackgroundFlusher:
    def __init__(
        self,
        buffer: "buffer_module.LogBuffer",
        http_client: "http_client_module.HTTPClient",
        rate_limiter: "rate_limiter_module.RateLimiter",
        flush_interval: float = 5.0,
        max_batch_size: int = 1000,
        max_retries: int = 3,
        retry_backoff_base: float = 2.0,
    ):
        self.buffer = buffer
        self.http_client = http_client
        self.rate_limiter = rate_limiter
        self.flush_interval = flush_interval
        self.max_batch_size = max_batch_size
        self.max_retries = max_retries
        self.retry_backoff_base = retry_backoff_base

        self._task: asyncio.Task[Any] | None = None
        self._shutdown_event = asyncio.Event()

        self._metrics = {
            "total_flushes": 0,
            "successful_flushes": 0,
            "failed_flushes": 0,
            "total_logs_sent": 0,
            "total_logs_failed": 0,
            "consecutive_failures": 0,
            "last_flush_time": None,
            "last_error": None,
            "errors_by_type": {},
        }

        self._circuit_breaker_open = False
        self._circuit_breaker_opened_at = None
        self._circuit_breaker_threshold = 5
        self._circuit_breaker_timeout = 60.0

    def start(self) -> None:
        if self._task is None or self._task.done():
            self._task = asyncio.create_task(self._run())

    async def _run(self) -> None:
        while not self._shutdown_event.is_set():
            try:
                await asyncio.sleep(self.flush_interval)

                if self.buffer.is_empty():
                    continue

                await self._flush_with_retry()

            except asyncio.CancelledError:
                break
            except Exception as e:
                self._log_error(f"Unexpected error in flusher: {e}")

    async def _flush_with_retry(self) -> None:
        if self._circuit_breaker_open:
            if time.time() - self._circuit_breaker_opened_at < self._circuit_breaker_timeout:
                return

            self._log_info("Circuit breaker: attempting recovery")
            self._circuit_breaker_open = False

        batch = await self.buffer.get_batch(self.max_batch_size)
        if not batch:
            return

        self._metrics["total_flushes"] += 1

        for attempt in range(self.max_retries):
            try:
                await self.rate_limiter.wait_if_needed()

                success = await self._send_batch(batch)

                if success:
                    self._metrics["successful_flushes"] += 1
                    self._metrics["total_logs_sent"] += len(batch)
                    self._metrics["consecutive_failures"] = 0
                    self._metrics["last_flush_time"] = time.time()
                    return

                if attempt < self.max_retries - 1:
                    backoff = self.retry_backoff_base**attempt
                    await asyncio.sleep(backoff)

            except httpx.TimeoutException as e:
                self._handle_network_error("Timeout", e, attempt)
                if attempt < self.max_retries - 1:
                    await asyncio.sleep(self.retry_backoff_base**attempt * 5.0)

            except httpx.ConnectError as e:
                self._handle_network_error("Connection refused", e, attempt)
                if attempt < self.max_retries - 1:
                    await asyncio.sleep(self.retry_backoff_base**attempt * 5.0)

            except Exception as e:
                self._handle_network_error("Unexpected error", e, attempt)
                if attempt < self.max_retries - 1:
                    await asyncio.sleep(self.retry_backoff_base**attempt)

        self._metrics["failed_flushes"] += 1
        self._metrics["total_logs_failed"] += len(batch)
        self._metrics["consecutive_failures"] += 1

        if self._metrics["consecutive_failures"] >= self._circuit_breaker_threshold:
            self._circuit_breaker_open = True
            self._circuit_breaker_opened_at = time.time()
            self._log_error(
                f"Circuit breaker OPEN: {self._metrics['consecutive_failures']} consecutive failures"
            )

    async def _send_batch(self, batch: list[dict[str, Any]]) -> bool:
        try:
            response = await self.http_client.post(
                "/api/v1/ingest/batch",
                json_data={"logs": batch},
            )

            if response.status_code == 202:
                data = response.json()
                accepted = data.get("accepted", 0)
                rejected = data.get("rejected", 0)

                if rejected > 0:
                    self._log_warning(f"Batch: {accepted} accepted, {rejected} rejected")
                    errors = data.get("errors", [])
                    for error in errors[:5]:
                        self._log_warning(f"  - {error}")

                return True

            if response.status_code == 429:
                retry_after = int(response.headers.get("Retry-After", 60))
                self._log_warning(f"Rate limit exceeded, sleeping {retry_after}s")
                self._increment_error_count("rate_limit")
                await asyncio.sleep(retry_after)
                return False

            if response.status_code == 503:
                retry_after = int(response.headers.get("Retry-After", 60))
                self._log_warning(f"Queue full (503), sleeping {retry_after}s")
                self._increment_error_count("queue_full")
                await asyncio.sleep(retry_after)
                return False

            if response.status_code == 401:
                self._log_error("Invalid API key (401), stopping ingestion")
                self._increment_error_count("auth_failure")
                self._shutdown_event.set()
                return False

            if response.status_code == 400:
                self._log_error(f"Bad request (400): {response.text}")
                self._increment_error_count("validation_error")
                return True

            self._log_error(f"Unexpected response: {response.status_code} - {response.text}")
            self._increment_error_count("server_error")
            return False

        except Exception:
            raise

    def _handle_network_error(self, error_type: str, error: Exception, attempt: int) -> None:
        self._log_error(f"{error_type} (attempt {attempt + 1}/{self.max_retries}): {error}")
        self._increment_error_count("network_error")
        self._metrics["last_error"] = f"{error_type}: {error}"

    def _increment_error_count(self, error_type: str) -> None:
        if error_type not in self._metrics["errors_by_type"]:
            self._metrics["errors_by_type"][error_type] = 0
        self._metrics["errors_by_type"][error_type] += 1

    def _log_info(self, message: str) -> None:
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        sys.stderr.write(f"[{timestamp}] [Ledger SDK] [INFO] {message}\n")
        sys.stderr.flush()

    def _log_warning(self, message: str) -> None:
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        sys.stderr.write(f"[{timestamp}] [Ledger SDK] [WARNING] {message}\n")
        sys.stderr.flush()

    def _log_error(self, message: str) -> None:
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        sys.stderr.write(f"[{timestamp}] [Ledger SDK] [ERROR] {message}\n")
        sys.stderr.flush()

    async def shutdown(self, timeout: float = 10.0) -> None:
        self._log_info("Shutting down, flushing remaining logs...")
        self._shutdown_event.set()

        if self._task and not self._task.done():
            try:
                await asyncio.wait_for(self._task, timeout=2.0)
            except asyncio.TimeoutError:
                self._task.cancel()
                try:
                    await self._task
                except asyncio.CancelledError:
                    pass

        remaining_attempts = 3
        while not self.buffer.is_empty() and remaining_attempts > 0:
            try:
                await asyncio.wait_for(self._flush_with_retry(), timeout=timeout / 3)
                remaining_attempts -= 1
            except asyncio.TimeoutError:
                self._log_warning("Flush timeout during shutdown")
                break
            except Exception as e:
                self._log_error(f"Shutdown flush error: {e}")
                break

        if not self.buffer.is_empty():
            dropped = self.buffer.size()
            self._log_warning(f"Shutdown: {dropped} logs still in buffer (not sent)")

        self._log_info("Shutdown complete")

    def get_metrics(self) -> dict[str, Any]:
        return {
            "total_flushes": self._metrics["total_flushes"],
            "successful_flushes": self._metrics["successful_flushes"],
            "failed_flushes": self._metrics["failed_flushes"],
            "total_logs_sent": self._metrics["total_logs_sent"],
            "total_logs_failed": self._metrics["total_logs_failed"],
            "consecutive_failures": self._metrics["consecutive_failures"],
            "circuit_breaker_open": self._circuit_breaker_open,
            "last_flush_time": self._metrics["last_flush_time"],
            "last_error": self._metrics["last_error"],
            "errors_by_type": self._metrics["errors_by_type"],
        }



import pytest

from ledger.core.buffer import LogBuffer


class TestLogBuffer:
    def test_buffer_initialization(self):
        buffer = LogBuffer(max_size=100)
        assert buffer.max_size == 100
        assert buffer.size() == 0
        assert buffer.is_empty() is True

    def test_add_log(self):
        buffer = LogBuffer(max_size=100)
        log_entry = {"message": "test", "level": "info"}

        buffer.add(log_entry)

        assert buffer.size() == 1
        assert buffer.is_empty() is False

    def test_buffer_overflow(self):
        buffer = LogBuffer(max_size=3)

        buffer.add({"id": 1})
        buffer.add({"id": 2})
        buffer.add({"id": 3})
        assert buffer.size() == 3
        assert buffer.get_dropped_count() == 0

        buffer.add({"id": 4})

        assert buffer.size() == 3
        assert buffer.get_dropped_count() == 1

    @pytest.mark.asyncio
    async def test_get_batch(self):
        buffer = LogBuffer(max_size=100)

        for i in range(10):
            buffer.add({"id": i})

        batch = await buffer.get_batch(5)

        assert len(batch) == 5
        assert buffer.size() == 5

    @pytest.mark.asyncio
    async def test_get_batch_empty_buffer(self):
        buffer = LogBuffer(max_size=100)

        batch = await buffer.get_batch(5)

        assert len(batch) == 0

    def test_clear_buffer(self):
        buffer = LogBuffer(max_size=100)

        buffer.add({"id": 1})
        buffer.add({"id": 2})
        assert buffer.size() == 2

        buffer.clear()

        assert buffer.size() == 0
        assert buffer.is_empty() is True



import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient

from ledger import LedgerClient
from ledger.integrations.fastapi import LedgerMiddleware


class TestFastAPIIntegration:
    @pytest.mark.asyncio
    async def test_middleware_logs_request(self, api_key, base_url):
        app = FastAPI()
        ledger = LedgerClient(api_key=api_key, base_url=base_url)

        app.add_middleware(
            LedgerMiddleware,
            ledger_client=ledger,
            exclude_paths=["/health"],
        )

        @app.get("/")
        async def root():
            return {"message": "Hello World"}

        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
            response = await client.get("/")

            assert response.status_code == 200
            assert response.json() == {"message": "Hello World"}

        await ledger.shutdown(timeout=0.1)

    @pytest.mark.asyncio
    async def test_middleware_excludes_paths(self, api_key, base_url):
        app = FastAPI()
        ledger = LedgerClient(api_key=api_key, base_url=base_url)

        app.add_middleware(
            LedgerMiddleware,
            ledger_client=ledger,
            exclude_paths=["/health"],
        )

        @app.get("/health")
        async def health():
            return {"status": "ok"}

        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
            response = await client.get("/health")

            assert response.status_code == 200

        await ledger.shutdown(timeout=0.1)

    @pytest.mark.asyncio
    async def test_middleware_logs_exception(self, api_key, base_url):
        app = FastAPI()
        ledger = LedgerClient(api_key=api_key, base_url=base_url)

        app.add_middleware(
            LedgerMiddleware,
            ledger_client=ledger,
        )

        @app.get("/error")
        async def error():
            raise ValueError("Test error")

        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
            with pytest.raises(ValueError):
                await client.get("/error")

        await ledger.shutdown(timeout=0.1)



import traceback
from datetime import datetime, timezone
from typing import Any

import ledger.core.buffer as buffer_module
import ledger.core.config as config_module
import ledger.core.flusher as flusher_module
import ledger.core.http_client as http_client_module
import ledger.core.rate_limiter as rate_limiter_module
import ledger.core.settings as settings_module
import ledger.core.validator as validator_module


class LedgerClient:
    def __init__(
        self,
        api_key: str,
        base_url: str | None = None,
        flush_interval: float | None = None,
        flush_size: int | None = None,
        max_buffer_size: int | None = None,
        http_timeout: float | None = None,
        http_pool_size: int | None = None,
        rate_limit_buffer: float | None = None,
    ):
        config = config_module.DEFAULT_CONFIG

        base_url = base_url or config.base_url
        flush_interval = flush_interval or config.flush_interval
        flush_size = flush_size or config.flush_size
        max_buffer_size = max_buffer_size or config.max_buffer_size
        http_timeout = http_timeout or config.http_timeout
        http_pool_size = http_pool_size or config.http_pool_size
        rate_limit_buffer = rate_limit_buffer or config.rate_limit_buffer
        self._validate_config(
            api_key=api_key,
            base_url=base_url,
            flush_interval=flush_interval,
            flush_size=flush_size,
            max_buffer_size=max_buffer_size,
            http_timeout=http_timeout,
            http_pool_size=http_pool_size,
            rate_limit_buffer=rate_limit_buffer,
        )

        self.api_key = api_key
        self.base_url = base_url

        self._http_client = http_client_module.HTTPClient(
            base_url=base_url,
            api_key=api_key,
            timeout=http_timeout,
            pool_size=http_pool_size,
        )

        self._settings_manager = settings_module.SettingsManager()

        self._buffer = buffer_module.LogBuffer(max_size=max_buffer_size)

        rate_limits = self._settings_manager.get_rate_limits()
        self._rate_limiter = rate_limiter_module.RateLimiter(
            requests_per_minute=rate_limits["requests_per_minute"],
            requests_per_hour=rate_limits["requests_per_hour"],
            buffer=rate_limit_buffer,
        )

        constraints = self._settings_manager.get_constraints()
        self._validator = validator_module.Validator(constraints)

        self._flusher = flusher_module.BackgroundFlusher(
            buffer=self._buffer,
            http_client=self._http_client,
            rate_limiter=self._rate_limiter,
            flush_interval=flush_interval,
            max_batch_size=constraints["max_batch_size"],
        )

        self._flusher.start()

        self._sdk_start_time = datetime.now(timezone.utc)

    def _validate_config(
        self,
        api_key: str,
        base_url: str,
        flush_interval: float,
        flush_size: int,
        max_buffer_size: int,
        http_timeout: float,
        http_pool_size: int,
        rate_limit_buffer: float,
    ) -> None:
        errors = []

        if not api_key or not isinstance(api_key, str):
            errors.append("api_key must be a non-empty string")

        if not api_key.startswith("ledger_"):
            errors.append("api_key must start with 'ledger_' prefix")

        if not base_url or not isinstance(base_url, str):
            errors.append("base_url must be a non-empty string")

        if not base_url.startswith(("http://", "https://")):
            errors.append("base_url must start with 'http://' or 'https://'")

        if flush_interval <= 0:
            errors.append(f"flush_interval must be positive, got {flush_interval}")

        if flush_size <= 0:
            errors.append(f"flush_size must be positive, got {flush_size}")

        if max_buffer_size <= 0:
            errors.append(f"max_buffer_size must be positive, got {max_buffer_size}")

        if http_timeout <= 0:
            errors.append(f"http_timeout must be positive, got {http_timeout}")

        if http_pool_size <= 0:
            errors.append(f"http_pool_size must be positive, got {http_pool_size}")

        if not 0 < rate_limit_buffer <= 1:
            errors.append(f"rate_limit_buffer must be between 0 and 1, got {rate_limit_buffer}")

        if errors:
            raise ValueError("Invalid Ledger SDK configuration:\n  - " + "\n  - ".join(errors))

    def log_info(
        self,
        message: str,
        attributes: dict[str, Any] | None = None,
    ) -> None:
        self._log(
            level="info",
            log_type="console",
            importance="standard",
            message=message,
            attributes=attributes,
        )

    def log_error(
        self,
        message: str,
        attributes: dict[str, Any] | None = None,
    ) -> None:
        self._log(
            level="error",
            log_type="console",
            importance="high",
            message=message,
            attributes=attributes,
        )

    def log_exception(
        self,
        exception: Exception,
        message: str | None = None,
        attributes: dict[str, Any] | None = None,
    ) -> None:
        stack_trace = "".join(
            traceback.format_exception(
                type(exception),
                exception,
                exception.__traceback__,
            )
        )

        self._log(
            level="error",
            log_type="exception",
            importance="high",
            message=message or str(exception),
            error_type=exception.__class__.__name__,
            error_message=str(exception),
            stack_trace=stack_trace,
            attributes=attributes,
        )

    def _log(
        self,
        level: str,
        log_type: str,
        importance: str,
        message: str | None = None,
        error_type: str | None = None,
        error_message: str | None = None,
        stack_trace: str | None = None,
        attributes: dict[str, Any] | None = None,
    ) -> None:
        from ledger._version import __version__

        log_entry: dict[str, Any] = {
            "timestamp": datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
            "level": level,
            "log_type": log_type,
            "importance": importance,
        }

        if message:
            log_entry["message"] = message

        if error_type:
            log_entry["error_type"] = error_type

        if error_message:
            log_entry["error_message"] = error_message

        if stack_trace:
            log_entry["stack_trace"] = stack_trace

        if attributes:
            log_entry["attributes"] = attributes

        log_entry["sdk_version"] = __version__
        log_entry["platform"] = "python"

        validated_log = self._validator.validate_log(log_entry)

        self._buffer.add(validated_log)

    def is_healthy(self) -> bool:
        flusher_metrics = self._flusher.get_metrics()

        if flusher_metrics["circuit_breaker_open"]:
            return False

        if flusher_metrics["consecutive_failures"] >= 3:
            return False

        buffer_utilization = (self._buffer.size() / self._buffer.max_size) * 100
        if buffer_utilization > 90:
            return False

        return True

    def get_health_status(self) -> dict[str, Any]:
        flusher_metrics = self._flusher.get_metrics()
        buffer_utilization = (self._buffer.size() / self._buffer.max_size) * 100

        status = "healthy"
        issues = []

        if flusher_metrics["circuit_breaker_open"]:
            status = "unhealthy"
            issues.append("Circuit breaker is open (too many failures)")

        if flusher_metrics["consecutive_failures"] >= 3:
            status = "degraded" if status == "healthy" else status
            issues.append(f"Consecutive failures: {flusher_metrics['consecutive_failures']}")

        if buffer_utilization > 90:
            status = "degraded" if status == "healthy" else status
            issues.append(f"Buffer nearly full: {buffer_utilization:.1f}%")

        if self._buffer.get_dropped_count() > 0:
            status = "degraded" if status == "healthy" else status
            issues.append(f"Dropped logs: {self._buffer.get_dropped_count()}")

        return {
            "status": status,
            "healthy": status == "healthy",
            "issues": issues if issues else None,
            "buffer_utilization_percent": round(buffer_utilization, 2),
            "circuit_breaker_open": flusher_metrics["circuit_breaker_open"],
            "consecutive_failures": flusher_metrics["consecutive_failures"],
        }

    async def shutdown(self, timeout: float = 10.0) -> None:
        await self._flusher.shutdown(timeout=timeout)
        await self._http_client.close()

    def get_metrics(self) -> dict[str, Any]:
        from ledger._version import __version__

        flusher_metrics = self._flusher.get_metrics()
        uptime = (datetime.now(timezone.utc) - self._sdk_start_time).total_seconds()

        return {
            "sdk": {
                "uptime_seconds": round(uptime, 2),
                "version": __version__,
            },
            "buffer": {
                "current_size": self._buffer.size(),
                "max_size": self._buffer.max_size,
                "total_dropped": self._buffer.get_dropped_count(),
                "utilization_percent": round(
                    (self._buffer.size() / self._buffer.max_size) * 100, 2
                ),
            },
            "flusher": {
                "total_flushes": flusher_metrics["total_flushes"],
                "successful_flushes": flusher_metrics["successful_flushes"],
                "failed_flushes": flusher_metrics["failed_flushes"],
                "total_logs_sent": flusher_metrics["total_logs_sent"],
                "total_logs_failed": flusher_metrics["total_logs_failed"],
                "consecutive_failures": flusher_metrics["consecutive_failures"],
                "circuit_breaker_open": flusher_metrics["circuit_breaker_open"],
                "last_flush_time": flusher_metrics["last_flush_time"],
                "last_error": flusher_metrics["last_error"],
            },
            "rate_limiter": {
                "current_rate": self._rate_limiter.get_current_rate(),
                "limit_per_minute": self._rate_limiter.limit_per_minute,
            },
            "errors": flusher_metrics["errors_by_type"],
        }



import asyncio
import sys
from collections import deque
from typing import Any


class LogBuffer:
    def __init__(self, max_size: int = 10000):
        self.max_size = max_size
        self._queue: deque[dict[str, Any]] = deque(maxlen=max_size)
        self._lock = asyncio.Lock()
        self._dropped_count = 0

    def add(self, log_entry: dict[str, Any]) -> None:
        if len(self._queue) >= self.max_size:
            self._queue.popleft()
            self._dropped_count += 1
            sys.stderr.write(
                f"[Ledger SDK] WARNING: Buffer full ({self.max_size} logs), "
                f"dropped oldest log (total dropped: {self._dropped_count})\n"
            )
            sys.stderr.flush()

        self._queue.append(log_entry)

    async def get_batch(self, max_batch_size: int) -> list[dict[str, Any]]:
        async with self._lock:
            batch_size = min(len(self._queue), max_batch_size)
            if batch_size == 0:
                return []

            batch = [self._queue.popleft() for _ in range(batch_size)]
            return batch

    def size(self) -> int:
        return len(self._queue)

    def is_empty(self) -> bool:
        return len(self._queue) == 0

    def clear(self) -> None:
        self._queue.clear()

    def get_dropped_count(self) -> int:
        return self._dropped_count
